
(* ### locations should be made abstract ### *)
type line = int;
type loc= line * int;	 (* type location (source line and char number), for type error messages *)
val L = (~1,~1);              (* the undefined location *)

(* for locating select alternants (used by parser) *)
val LOCATIONS = ref ([] : loc list);
fun pushLocation l = LOCATIONS := l :: !LOCATIONS;
fun popLocations k =
    let val locs = List.take (!LOCATIONS,k)
    in LOCATIONS := List.drop (!LOCATIONS,k);
       isort locs (fn ((l1,c1),(l2,c2)) => l1>l2 orelse (l1=l2 andalso c1 > c2)) 
    end
    handle _ => error "Parsing error (popLocations), please report it";
(* for debug, prints a location list:
fun plocs locs =
    let open ErrPrint
	fun printp (x,y) = (print "("; printi x; print "."; printi y; print ")")
    in print  "locations = "; print_list  ("[",", ","]\n") (fn l => print_list ("[",", ","]") printp (List.map fst l)) locs
    end;
*)
fun zipSelect sll =
    let val k = List.length (List.concat sll)
	val starts = popLocations k
	fun zip [] _ = []
	  |  zip (sl::sll) ll =
	     let val z = List.length sl
		 val l = List.take (ll,z)
	     in ListPair.zipEq (l,sl) :: zip sll (List.drop (ll,z))
	     end
    in zip sll starts
    end;
fun zipMatches ml =
    let val k = List.length ml
	val starts = popLocations k
    in List.map (fn (loc,(r,p,l,s)) => (loc,r,p,l,s)) (ListPair.zipEq (starts,ml))
    end;



structure Fiacre = struct

datatype kind =							       (* kinds of idents in exprs	*)
	  OpenKind							       (* undetermined			*)
	| VarKind							               (* variable			*)
	| ConstantKind						       (* constant and value		*)
	| ConstrKind							       (* constructor			*)
        | FfunKind                                                               (* fiacre function              *)
        | CfunKind of string;                                               (* C function with cname        *)

type  Name = loc * string;                                            (* located string *)

datatype Fiacre =
	 Program of Declaration list		                        (* declarations, including main *)
and Declaration =
	  TypeDecl of loc * Name * Type				(* decl location, name, body		*)
	| ChannelDecl of loc * Name * Channel			(* location, name, body		*)
	| ConstantDecl of loc * Name * Type * Exp		(* location, name, type, body	*)
	| FunctionDecl of
	        loc *							        (* location of "function"	*)
		Name *						                (* name			        *)
		FunParamDecl list *                                        (* arguments                    *)
		Type *                                                             (* result type                  *)
		VarDecl list *						        (* local variables		*)
		Statement                                                  	(* body                         *)
	| ExternDecl of
	        loc *								(* location			*)
		Name *					                	(* name			        *)
		Type list *                                               	(* argument types               *)
		Type *                                                     	(* result type                  *)
		string                                                       	(* C symbol                     *)
	| ProcessDecl of
		loc *								(* location			*)
		Name *						        	(* name				*)
		PortDecl list *							(* ports parameters		*)
		ParamDecl list *						(* value or ref parameters	*)
		StatesDecl *							(* states			*)
		VarDecl list *							(* local variables		*)
		(loc * Statement) option *				(* optional init statement	*)
		Transition list							(* the list of transitions	*)
	| ComponentDecl of
		loc *								(* location			*)
		Name *						        	(* name				*)
		PortDecl list *							(* ports parameters		*)
		ParamDecl list *						(* value or ref parameters	*)
		VarDecl list *							(* local variables		*)
		locPortDecl list *						(* hidden ports and delays	*)
		PriorDecl list *						(* local priority relation	*)
		(loc * Statement) option *				(* optional init statement	*)
		Composition							(* body				*)
	| Main of Name * string list                                    (* program main and prio rel.   *)
	| PropertyDecl of loc * Name * Property		(* location, name, prop body      *)
	| Assert of           	       			   	     	   	(* assert property              *)
	        loc *
		Name *
		string option *
		string option
and Arg =
	  ValArg of Name							(* arg passed by value		*)
	| RefArg of Name							(* arg passed by reference	*)
and Type =
	  BoolType								(* bool				*)
	| NatType								(* nat				*)
	| IntType							           	(* int				*)
	| NamedType of string						(* abbreviation			*)
	| IntervalType of Exp * Exp			        	(* interval exp..exp		*)
	| UnionType of (string list * Type) list	        	(* constructors, BotType or type *)
	| RecordType of (string list * Type) list			(* record			*)
	| ArrayType of Exp * Type					(* array exp of type		*)
	| QueueType of Exp * Type					(* queue exp of type		*)
	| AnyArray							   	(* top array type, internal	*)
	| AnyQueue							   	(* top queue type, internal	*)
	| BotType							   	(* bottom type, internal	*)
	| TopType							   	(* top type, internal		*)
and Exp = 
  	  ParenExp of Exp							(* (exp), kept for tracability	*)
	| CondExp of Exp * Exp * Exp * Type ref        	(* conditional exp              *)
  	| InfixExp of Infix * Type ref * Exp * Exp		(* infix operators		*)
  	| BinopExp of Binop * Type ref * Exp * Exp		(* binary operators		*)
  	| UnopExp of Unop * Type ref * Exp			(* unary operators		*)
	| RecordExp of (string * Exp) list * Type ref		(* {f1=e1, ..., fn=vn}		*)
	| QueueExp of Exp list * Type ref			   	(* queue constants, type	*)
	| ArrayExp of Exp list * Type ref				(* [e1, ..., en]		*)
        | RefExp of string * Type ref                  		(* only in instance arguments	*)
	| ConstrExp of string * Exp * Type ref			(* constructions		*)
	| IntExp of int * Type ref						(* numeric literal, type	*)
	| BoolExp of bool							(* bool literal			*)
	| IdentExp of string * (kind * Type) ref			(* var, constant or constructor	*)
	| AnyExp							        	(* "any" pattern		*)
	| ArrayAccessExp of Exp * Exp * Type ref		(* array subscript		*)
	| RecordAccessExp of Exp * string * Type ref	(* record field access		*)
	| JuxtExp of Exp * Exp list * Type list ref        	(* juxtaposition                *)
and Infix =
  	  AND | OR | IMP | LE | LT | GT | GE | EQ | NE | ADD | SUB | MOD | MUL | DIV
and Binop =
	  ENQUEUE | APPEND
and Unop =
  	  MINUS	| PLUS | NOT | ASSERT of int | COERCE | FULL | EMPTY | DEQUEUE	| FIRST | LENGTH
and Statement =
	  StNull of loc								(* does nothing statement	*)
	| StCond of					                	(* if e1 then s1 [elsif ei then si] [else sn] 	*)
		loc *								(* if location			*)
		Exp *								(* condition				*)
		(loc * Statement) *						(* then location and st			*)
		(loc * Exp * Statement) list *				(* elsif locations and st 		*)
		(loc * Statement) option *				(* else location and st  	*)
		loc                                                           	(* end location  *)
	| StSelect of	                                          		(* select s1 [] .. sn (unless ...) ... end *)
	        (loc *                                                       	(* select or unless or [] location *)
		 Statement) list list *                             		(* st *)
		loc                                                           	(* end location *)
	| StSequence of	                                       		(* s1; s2			*)
	        loc *                                                         	(* ";" location *)
		Statement *
		Statement
	| StTo of									(* to state, loop *)
	        loc *                                                       		(* to location *)
		string                                                     		(* state or "" (meaning loop) *)
	| StWhile of				                       		(* while e do s end *)
	        loc *                                                        	(* while location *)
		Exp *                                                       	(* condition *)
		loc *                                                        	(* do location *)
		Statement *                                           		(* st *)
		loc                                                           	(* end location *)
	| StAssign of 		                                  		(* access list := exp list	*)
	        loc *                                                         	(* ":=" location *)
		Pattern list *
		Exp list
	| StAny of								(* access list:=any [where exp]	*)
	         loc *                                                        	(* ":=" location *)
		 Pattern list *
		 (Type list) ref *
		 Exp option
	| StSignal of Name * Profile ref				(* event			*)
	| StInput of Name * Profile ref * Pattern list * Exp option  	   (* p?access list [where exp] *)
	| StOutput of Name * Profile ref * Exp list         	(* p!exp list			*)
	| StCase of                                                  		(* case exp of matches	*)
	        loc *                                                          	(* case location	*)
		Exp *                                                         	(* exp 	*)
		(loc *                                                        	(* of or "|" location	*)
		 bool ref *                                                 	(* records redundancy information *)
		 Pattern *                                                  	(* pattern *)
		 loc *                                                         	(* "->" location *)
		 Statement) list *                                      	(* st *)
		loc                                                                   (* end location *)
	| StForeach of loc * string * Type ref * loc * Statement * loc     (* foreach v do ... end         *)
        | StOn of loc * Exp                                         	      	(* test e                       *)
        | StWait of loc * Bound * Bound                  		(* time interval                *)
        | StReturn of loc * Exp                                        	(* return from function         *)
	| StTag of loc * string                                        	(* for tag probes               *)
and Bound = 
	  Open of real								(* ]from or to[			*)
	| Closed of real							(* [from or to]			*)
	| Infinite								   	(* ...				*)
and VarAttr =
	  READ | WRITE							(* shared variables attributes  *)
and PortAttr =
	  IN | OUT
and Composition =
	  ParComp of loc * PortSet * (PortSet * Composition) list	(* par set_0 in set_i -> comp_i	*)
	| InstanceComp of Instance					(* name [pi] (ei or &v)		*)
and PortSet =
	  AllPorts									(* sort ( * )			*)
	| SomePorts of string list					(* port list, possibly empty	*)
and Probe = (* observables *)
          StateProbe of Path * string
        | ValueProbe of Path * Exp
        | EventProbe of Path * string
        | TagProbe of Path * string
        | EnterProbe of Probe
        | LeaveProbe of Probe
	| AndProbe of Probe * Probe
	| OrProbe of Probe * Probe
	| ImpProbe of Probe * Probe
	| NotProbe of Probe
        | ParenProbe of Probe
        (* as parsed: *)
	| JuxtProbe of string * Probe
	| ThatProbe of string * Path * Exp
and Property =
         DeadlockFreeProp											(* [] - dead                                     *)
       | InfinitelyOftenProp of Probe                                                			(* [] <> p                                       *)
       | MortalProp of Probe                                                         			(* <> [] - p                                     *)
       | LiveProp of Probe                                                           			(* AG EF p     (currently rejected by parser)    *)
       (* presence *)
       | PresentProp of Probe * Interval option							(* present A {within I}                          *)
       | PresentLastingProp of Probe * Bound                                         		(* present A lasting D                           *)
       | PresentAfterProp of Probe * Probe * Interval option                         	(* present A after B {within I}                  *)
       | PresentBeforeProp of Probe * Probe * Interval option                        	(* present {first} A before B {within I}         *)
       | PresentBetweenProp of Probe * Probe * Probe * Interval option        	(* present A between B and C {within I}          *)
       | PresentUntilProp of Probe * Probe * Probe * Interval option           	(* present A after B until C {within I}          *)
       (* absence *)
       | AbsentProp of Probe * Interval option							(* absent A {within I}                           *)
       | AbsentLastingProp of Probe * Bound                                          		(* absent A lasting D                            *)
       | AbsentAfterProp of Probe * Probe * Interval option                    		(* absent A after B {within I}                   *)
       | AbsentBeforeProp of Probe * Probe * Bound option                    		(* absent A before B {for duration D}            *)
       | AbsentBetweenProp of Probe * Probe * Probe * Interval option          (* absent A between B and C {within I}           *)
       | AbsentUntilProp of Probe * Probe * Probe * Interval option            	(* absent A after B until C {within I}           *)
       (* response *)
       | LeadstoProp of Probe * Probe * Interval option					(* A leadsto {first} B {within I}                *)
       | LeadstoBeforeProp of Probe * Probe * Probe * Interval option            (* A leadsto {first} B {within I} before C       *)
       | LeadstoAfterProp of Probe * Probe * Probe * Interval option               (* A leadsto {first} B {within I} after C        *)
       | LeadstoBetweenProp of Probe * Probe * Probe * Probe * Interval option       (* A leadsto B between Q and R {within I}        *)
       | LeadstoUntilProp of Probe * Probe * Probe * Probe *Interval option    	   (* A leadsto B after Q until R {within I}        *)
       (* universality *)
       | AlwaysProp of Probe * Interval option							(* always A {within I}                           *)
       | AlwaysLastingProp of Probe * Bound                                          		(* always A lasting D                            *)
       | AlwaysAfterProp of Probe * Probe * Bound option                             	(* always A after B {for duration D}             *)
       | AlwaysBeforeProp of Probe * Probe * Bound option                     	(* always A before B {for duration D}            *)
       | AlwaysBetweenProp of Probe * Probe * Probe * Interval option   	(* always A between B and C {within I}           *)
       | AlwaysUntilProp of Probe * Probe * Probe * Interval option           	(* always A after B until C {within I}           *)
       (* precedence *)
       | PrecedesProp of Probe * Probe * Bound option					(* A precedes B {for duration D}                 *)
       | PrecedesBeforeProp of Probe * Probe * Probe * Bound option     	(* A precedes B before R {for duration D}        *)
       | PrecedesAfterProp of Probe * Probe * Probe * Bound option        	(* A precedes B after R {for duration D}         *)
       | PrecedesBetweenProp of Probe * Probe * Probe * Probe * Bound option         (* A precedes B between Q and R {for duration D} *)
       | PrecedesUntilProp of Probe * Probe * Probe * Probe * Bound option   	     (* A precedes B after Q until R {for duration D} *)
       (* composite *)
       | AndProp of Property * Property
       | OrProp of Property * Property
       | ImpProp of Property * Property
       | NotProp of Property
       (* ltl props *)
       | BoxProp of Property
       | DiaProp of Property
       | NextProp of Property
       | LtlProp of Property
       | ProbeProp of Probe
       | UntilProp of Property * Property
       | ReleaseProp of Property * Property
       (* parenthesis *)
       | ParenProp of Property
       (* variable *)
       | VarProp of loc * string

withtype
	Profile = PortAttr list * Type list
and  Channel = Type list
and 	ChannelDecl = loc * string * (* Channel *) Type list
and	PortDecl = Name list * (* Profile *) (PortAttr list * Type list)
and	FunParamDecl = Name list * Type                      (* no & nor var attributes yet *)
and	ParamDecl = Arg list * VarAttr list * Type
and	StatesDecl = Name list
and	VarDecl = Name list * Type * Exp option
and 	PriorDecl = Name list * Name list
and 	locPortDecl = (Name list * (* Profile *) (PortAttr list * Type list)) * Probe option * (Bound * Bound)
and  Instance = Name * string list * Exp list
and	Transition = loc * Name * Statement
and  Literal = Exp
and  Pattern = Exp
and  Interval = Bound * Bound
and  Path = string * int list
;

end;


open Fiacre;


(* sorts record fields in record expressions *)
fun sortvfields fl = isort fl (fn ((f1,_),(f2,_)) => String.< (f2,f1));

(* propagates unary - or + to enclosed literal, if any, so that negative integers
   can appear in patterns *)
fun normPlus (UnopExp(_,_,a as IntExp _)) = a | normPlus e = e;
fun normMinus (UnopExp(_,_,IntExp (k,rty))) = IntExp (~k,rty) | normMinus e = e;

(* for pairs of bounds encoding intervals *)
fun nontrivial (Closed x,Infinite) = x > 0.0 | nontrivial _ = true;


(* EXPVAL filled by typer, needed to compute lengths in queues and arrays in frac json printer *)
val EXPVAL = ref (fn e : Exp => e);
